/*
 * Copyright (c) 2003 Sergey Zaharchenko <doublef@nm.ru>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY SERGEY ZAHARCHENKO ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE SERGEY ZAHARCHENKO BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include <X11/Xlib.h>

#ifdef WITHOUT_IMLIB
#include <X11/xpm.h>
#include <X11/extensions/shape.h>
#else
#include <Imlib.h>
#endif

#include <sys/socket.h>
#include <sys/fcntl.h>
#include <signal.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>

Display*Scr;
Window	W;

XSetWindowAttributes	WAttrs;

#ifndef WITHOUT_IMLIB
typedef ImlibImage pic;
#else
typedef struct
{
	Pixmap  Picture;
	Pixmap  Shape;
	int     rgb_width;
	int     rgb_height;
} pic;

XpmAttributes		Attrs;

#endif

int	shX=0,shY=0;

FILE*	F;
int	f;

char*	getstring(FILE* F)
{
	char*	result=(char*)malloc(1);
	int	pos=0;
	int	q;

	while ((q=fgetc(F))!='\n')
	{
		result[pos]=q;
		pos++;
		result=(char*)realloc(result,pos+1);
	}
	result[pos]='\0';
	clearerr(F);
	if (pos==0)
	{
		free(result);
		result=NULL;
	}
	return result;
}

void	quit(int sig)
{
	fprintf(F,"quit\n");
	fflush(F);
}

int	main(int argc, char** argv)
{
	XEvent		Event;
	char*		str;
	char		leave=0;
	fd_set		theSockets;
#ifndef	WITHOUT_IMLIB
	ImlibData*	data;
#else
#endif
	Scr=XOpenDisplay(NULL);

#ifndef	WITHOUT_IMLIB
	data=Imlib_init(Scr);
#endif

	W=XCreateSimpleWindow(Scr,RootWindow(Scr, 0), 0, 0, 1, 1, 0,
		BlackPixel(Scr, 0), BlackPixel(Scr, 0));
	
	WAttrs.backing_store=Always;
	WAttrs.override_redirect=True;

	XChangeWindowAttributes(Scr, W, CWBackingStore|CWOverrideRedirect,
		&WAttrs);

	XSelectInput(Scr,W, KeyPressMask|KeyReleaseMask|
		ButtonPressMask|ButtonReleaseMask|
		EnterWindowMask|LeaveWindowMask|PointerMotionMask);

	signal(SIGTERM,quit);
	signal(SIGHUP,quit);
	signal(SIGPIPE,quit);

	F=popen(argv[1],"r+");
	f=fileno(F);

	do
	{
		FD_ZERO(&theSockets);
		FD_SET(f,&theSockets);
		FD_SET(ConnectionNumber(Scr),&theSockets);
		select(FD_SETSIZE,&theSockets,NULL,NULL,NULL);

		if (FD_ISSET(f,&theSockets))
		if ((str=getstring(F))!=NULL)
		{
			char*	command=(char*)malloc(strlen(str)+1);
			int	compos=0;
			printf("%s\n",str);
			sscanf(str,"%s%n",command,&compos);
			compos++;
			if (!strcmp(command,"pid"))
			{
				fprintf(F,"%d\n",getpid());
				fflush(F);
			} else
			if (!strcmp(command,"dimen"))
			{
				Window root;
				int	x,y,dx,dy,bw,depth;

				XGetGeometry(Scr,RootWindow(Scr,0),
					&root,&x,&y,&dx,&dy,&bw,&depth);

				fprintf(F,"%d %d\n",dx,dy);
				fflush(F);
			} else
			if (!strcmp(command,"load"))
			{
				char* fn=(char*)malloc(strlen(str+compos)+1);
				pic * p;
				sscanf(str+compos,"%s",fn);
				printf("%s\n",fn);
#ifndef WITHOUT_IMLIB
				p=(pic*)Imlib_load_image(data,fn);
#else
				p=(pic*)malloc(sizeof(pic));
				XpmReadFileToPixmap(Scr, RootWindow(Scr, 0), fn,
					&p->Picture, &p->Shape, &Attrs);
				XSync(Scr,0);
				p->rgb_width=Attrs.width;
				p->rgb_height=Attrs.height;

#endif
				fprintf(F,"%p %d %d\n",
					p,p->rgb_width,p->rgb_height);

				fflush(F);
				free(fn);
			} else
			if (!strcmp(command,"display"))
			{
				pic * p;
				sscanf(str+compos,"%p%i%i",&p,&shX,&shY);

				XMoveResizeWindow(Scr, W,
					shX, shY, p->rgb_height, p->rgb_width);
#ifndef WITHOUT_IMLIB
				Imlib_apply_image(data,p,W);
#else
				XSetWindowBackgroundPixmap(Scr, W,
					p->Picture);
				XShapeCombineMask(Scr,W,ShapeBounding,0,0,
					p->Shape,ShapeSet);
				XClearWindow(Scr, W);
#endif
				XMapWindow(Scr, W);
				XSync(Scr,0);
			} else
			if (!strcmp(command,"unload"))
			{
				pic * p;
				sscanf(str+compos,"%p",&p);
#ifndef WITHOUT_IMLIB
				Imlib_destroy_image(data,p);
#else
				XFreePixmap(Scr,p->Picture);
				XFreePixmap(Scr,p->Shape);
				free(p);
#endif
			} else
			if (!strcmp(command,"warn"))
			{
				printf("%s\n",str+compos);
			} else
			if (!strcmp(command,"quit"))
			{
				leave=1;
			};
			free(command);
			free(str);
		}

		if (FD_ISSET(ConnectionNumber(Scr),&theSockets))
		if (XEventsQueued(Scr,QueuedAfterFlush)>0)
		{
			XNextEvent(Scr, &Event);
			if (Event.xany.window==W)
			{
				switch (Event.type)
				{
				case KeyPress:
					fprintf(F,"key press %d %d\n",
						Event.xkey.state,
						Event.xkey.keycode);

					fflush(F);
				break;
				case KeyRelease:
					fprintf(F,"key release %d %d\n",
						Event.xkey.state,
						Event.xkey.keycode);

					fflush(F);
				break;
				case ButtonPress:
					fprintf(F,"button press %d %d %d\n",
						Event.xbutton.button,
						Event.xbutton.x,
						Event.xbutton.y);

					fflush(F);
				break;
				case ButtonRelease:
					fprintf(F,"button release %d %d %d\n",
						Event.xbutton.button,
						Event.xbutton.x,
						Event.xbutton.y);

					fflush(F);
				break;
				case EnterNotify:
					fprintf(F,"enter\n");
					fflush(F);
				break;
				case LeaveNotify:
					fprintf(F,"leave\n");
					fflush(F);
				break;
				case MotionNotify:
					fprintf(F,"motion %d %d %d\n",
						Event.xmotion.x,
						Event.xmotion.y,
						Event.xmotion.state);

					fflush(F);
				break;
				};
			};
		};
	} while (!leave);

	signal(SIGTERM,SIG_DFL);
	signal(SIGHUP,SIG_DFL);
	signal(SIGPIPE,SIG_DFL);

	return 0;
}
